%{
#include <stdlib.h>
#include <string.h>
#include "nsu.h"
#include "nsu_app.h"

#define MAX_RR 128

static nsu_rr rr_base[MAX_RR];
static int rr_count = 0;
static int update_start = -1;

static nsu_rr *
alloc_rr(int op, char *name)
{
	nsu_rr *rr;
	
	if (rr_count == MAX_RR)
		abend("too many RRs");
	if (strlen(name) + 1 > sizeof(rr->name))
		abend("%s: name too long", name);
	rr = &rr_base[rr_count++];
	memset(rr, 0, sizeof *rr);
	rr->op = op;
	strcpy(rr->name, name);
	return rr;
}

int rdata_parse(nsu_rr *rr, int type);

%}

%token T_NXDOMAIN
%token T_YXDOMAIN
%token T_NXRRSET
%token T_YXRRSET
%token T_ADD
%token T_DELETE
%token T_SEND
%token T_UPDATE
%token <s> T_STRING
%token <n> T_NUMBER

%type <s> domain
%type <n> ttl type
%type <r> rdata

%union {
	char *s;
	unsigned long n;
	nsu_rr r;
}

%%

input       : stmt
            | input stmt
            ;

stmt        : opt_prereq_list { update_start = rr_count; } update_stmt
              {
		      run_update(rr_base, update_start,
				 rr_base + update_start,
				 rr_count - update_start);
		      dealloc_strings();
		      rr_count = 0;
	      }
            | ';'
            | T_SEND ';'
	    ;

opt_prereq_list : /* empty */
            | prereq_list
	    ;

prereq_list : prereq 
            | prereq_list prereq
            ;

prereq      : prereq_cond ';'
            ;

prereq_cond : T_NXDOMAIN domain
              {
		      alloc_rr(NSU_PREREQ_NXDOMAIN, $2);
	      }
            | T_YXDOMAIN domain
              {
		      alloc_rr(NSU_PREREQ_YXDOMAIN, $2);
	      }
            | T_NXRRSET domain type
	      {
		      nsu_rr *rr = alloc_rr(NSU_PREREQ_NXRRSET, $2);
		      rr->type = $3;
	      }
            | T_YXRRSET domain type rdata
	      {
		      nsu_rr *rr = alloc_rr(NSU_PREREQ_YXRRSET, $2);
		      rr->type = $3;
		      if ($4.type == $3)
			      memcpy(&rr->data, &$4.data, sizeof(rr->data));
	      }
            ;

rdata       : /* emtpy */
              {
		      rdata_parse(&$$, $<n>0);
	      }
            ;

update_stmt : update
            | update_stmt update
            ;

update      : update_expr ';'
            | T_UPDATE update_expr ';'
            ;

update_expr : T_DELETE domain
              {
		      alloc_rr(NSU_UPDATE_DELALL, $2);
	      }
            | T_DELETE domain type rdata
	      {
		      nsu_rr *rr = alloc_rr($3 == $4.type ?
					    NSU_UPDATE_DELV : NSU_UPDATE_DEL,
					    $2);
		      rr->type = $3;
		      if (rr->op == NSU_UPDATE_DELV)
			      memcpy(&rr->data, &$4.data, sizeof(rr->data));
	      }
            | T_ADD domain ttl type rdata
	      {
		      nsu_rr *rr;
		      
		      if ($4 != $5.type) {
			      yyerror("need RDATA");
			      YYERROR;
		      }
		      rr = alloc_rr(NSU_UPDATE_ADD, $2);
		      rr->ttl = $3;
		      rr->type = $4;
		      memcpy(&rr->data, &$5.data, sizeof(rr->data));
	      }
            ;

domain      : { need_domain(); } T_STRING
              {
		      $$ = $2;
	      }
            ;

ttl         : /* empty */
              {
		      $$ = default_ttl; //FIXME
	      }
            | T_NUMBER
	    ;

type        : T_STRING
              {
		      $$ = nsu_str_to_type($1);
		      if ($$ == ns_t_invalid)
			      abend("%s: invalid or unsupported type", $1);
              }
            ;
%%
int
yyerror(const char *msg)
{
	abend("%s", msg);
}

